/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.CommandLineTool;
import info.yeppp.Build.internals.FileSystem;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.IOException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MicrosoftLibraryManager extends AbstractLibrarian {

	public MicrosoftLibraryManager(Architecture architecture, VisualStudio visualStudio) {
		this.architecture = architecture;
		this.visualStudio = visualStudio;
		this.librarianPath = FilenameUtils.concat(visualStudio.getCppToolsPath(architecture), "lib.exe");
	}

	public BuildMessages createStaticLibrary(String libraryPath, Collection<String> objectPaths) {
		if (binariesDirectory == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();
		try {
			final List<String> commandLine = composeCommandLine(libraryPath + ".lib", objectPaths);
			final Map<String, String> environment = composeEnvironment();
			final CommandLineTool commandLineTool = new CommandLineTool(commandLine, environment);
			assert commandLineTool.getError().isEmpty();
			buildMessages.add(parseCommandLineOutput(commandLineTool.getOutput(), libraryPath, FileSystem.getRelativePath(libraryPath, binariesDirectory)));
		} catch (Exception exception) {
			buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, libraryPath, null, null, exception.getMessage()));
		}
		return buildMessages;
	}

	protected BuildMessages parseCommandLineOutput(List<String> lines, String sourcePath, String relativeSourcePath) {
		final BuildMessages buildMessages = new BuildMessages();
		final Pattern fatalErrorPattern = Pattern.compile("^LINK : fatal error LNK(\\d+): (.+)$");
		final Pattern warningPattern = Pattern.compile("^(.+) : warning LNK(\\d+): (.+)$");
		Matcher matcher;
		for (String line : lines) {
			if ((matcher = fatalErrorPattern.matcher(line)).find()) {
				final int errorNumber = Integer.parseInt(matcher.group(1));
				final String errorMessage = matcher.group(2);
				buildMessages.add(new BuildMessage(BuildMessage.Type.FatalError, null, null, errorNumber, errorMessage));
			} else if ((matcher = warningPattern.matcher(line)).find()) {
				final String errorPath = matcher.group(1);
				final int errorNumber = Integer.parseInt(matcher.group(2));
				final String errorMessage = matcher.group(3);
				buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, errorPath, null, errorNumber, errorMessage));
			} else {
				final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
				if (lastBuildMessage == null) {
					buildMessages.add(new BuildMessage(BuildMessage.Type.Error, null, null, null, line));
				} else {
					lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
				}
			}
		}
		return buildMessages;
	}

	private List<String> composeCommandLine(String outputPath, Collection<String> objectPaths) {
		final List<String> commandLine = new ArrayList<String>();
		commandLine.add(librarianPath);
		commandLine.add("/nologo");
		for (String libraryDirectory : libraryDirectories) {
			commandLine.add("/LIBPATH:" + libraryDirectory);
		}
		switch (architecture) {
			case X86:
				commandLine.add("/MACHINE:X86"); break;
			case X64:
				commandLine.add("/MACHINE:X64"); break;
			case IA64:
				commandLine.add("/MACHINE:IA64"); break;
			case ARM:
				commandLine.add("/MACHINE:ARM"); break;
		}
		commandLine.add("/OUT:" + outputPath);
		for (String objectPath : objectPaths) {
			commandLine.add(objectPath);
		}
		for (String library : libraries) {
			commandLine.add(library + ".lib");
		}
		return commandLine;
	}

	private Map<String, String> composeEnvironment() throws IOException {
		final Map<String, String> environment = new HashMap<String, String>(System.getenv());
		boolean processedIncludeVariable = false;
		for (Map.Entry<String, String> variable : environment.entrySet()) {
			if (variable.getKey().equalsIgnoreCase("PATH")) {
				String path = variable.getValue();
				path = visualStudio.getIdePath() +
						SystemUtils.PATH_SEPARATOR + visualStudio.getCommonToolsPath() +
						SystemUtils.PATH_SEPARATOR + visualStudio.getCppToolsPath(architecture) +
						SystemUtils.PATH_SEPARATOR + path;
				variable.setValue(path);
			} else if (variable.getKey().equalsIgnoreCase("LIB")) {
				String include = variable.getValue();
				include = StringUtils.join(globalLibraryDirectories, SystemUtils.PATH_SEPARATOR) +
						SystemUtils.PATH_SEPARATOR + include;
				variable.setValue(include);
				processedIncludeVariable = true;
			}
		}
		if (processedIncludeVariable == false) {
			environment.put("LIB", StringUtils.join(globalLibraryDirectories, SystemUtils.PATH_SEPARATOR));
		}
		return environment;
	}

	@Override
	public void addDefaultGlobalLibraryDirectories() {
		this.addGlobalLibraryDirectory(visualStudio.getCppLibraryPath(this.architecture));
		this.addGlobalLibraryDirectory(visualStudio.windowsSdk.getLibraryDirectory(this.architecture));
	}

	private final VisualStudio visualStudio;
	private final Architecture architecture;
	private final String librarianPath;

}
