/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.*;
import info.yeppp.Build.internals.Process;
import info.yeppp.Build.sdk.WindowsSdk;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MicrosoftLinker extends AbstractLinker {

	public MicrosoftLinker(VisualStudio visualStudio, WindowsSdk windowsSdk, Architecture architecture) {
		this.architecture = architecture;
		this.visualStudio = visualStudio;
		this.windowsSdk = windowsSdk;
		this.linkerPath = FilenameUtils.concat(visualStudio.getCppToolsPath(architecture), "link.exe");
	}

	public MicrosoftLinker(MicrosoftLinker prototype) {
		super(prototype);
		architecture = prototype.architecture;
		visualStudio = prototype.visualStudio;
		windowsSdk = prototype.windowsSdk;
		linkerPath = prototype.linkerPath;
	}

	public BuildMessages linkExecutable(final String executablePath, Collection<String> objectPaths) {
		if (binariesDirectory == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		FileUtils.deleteQuietly(new File(executablePath + ".exe"));

		final info.yeppp.Build.internals.Process.Parameters parameters = prepareParameters(executablePath + ".exe", objectPaths, false, null);
		final Process.Environment environment = prepareEnvironment();
		final Process.EventHandler eventHandler = new Process.EventHandler() {
			final Pattern fatalErrorPattern = Pattern.compile("^LINK : fatal error LNK(\\d+): (.+)$");
			final Pattern warningPattern = Pattern.compile("^(.+) : warning LNK(\\d+): (.+)$");

			@Override
			public void onStdOutLineWrite(String line) {
				System.out.println(line);
			}

			@Override
			public void onStdErrLineWrite(String line) {
				Matcher matcher;
				if ((matcher = fatalErrorPattern.matcher(line)).find()) {
					final int errorNumber = Integer.parseInt(matcher.group(1));
					final String errorMessage = matcher.group(2);
					buildMessages.add(new BuildMessage(BuildMessage.Type.FatalError, null, null, errorNumber, errorMessage));
				} else if ((matcher = warningPattern.matcher(line)).find()) {
					final String errorPath = matcher.group(1);
					final int errorNumber = Integer.parseInt(matcher.group(2));
					final String errorMessage = matcher.group(3);
					buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, errorPath, null, errorNumber, errorMessage));
				} else {
					final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
					if (lastBuildMessage == null) {
						buildMessages.add(new BuildMessage(BuildMessage.Type.Error, null, null, null, line));
					} else {
						lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
					}
				}
			}

			@Override
			public void onError(String description) {
				System.err.println(description);
				buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, executablePath, null, null, description));
			}

			@Override
			public void onCompletion(Integer status) {
			}
		};

		final Process linker = new Process(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	@Override
	public BuildMessages linkDynamicLibrary(final String libraryPath, Collection<String> objectPaths, String specificationPath) {
		if (binariesDirectory == null) {
			throw new IllegalStateException("Binaries directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();

		FileUtils.deleteQuietly(new File(libraryPath + ".dll"));

		final Process.Parameters parameters = prepareParameters(libraryPath + ".dll", objectPaths, true, specificationPath);
		final Process.Environment environment = prepareEnvironment();
		final Process.EventHandler eventHandler = new Process.EventHandler() {
			final Pattern fatalErrorPattern = Pattern.compile("^LINK : fatal error LNK(\\d+): (.+)$");
			final Pattern warningPattern = Pattern.compile("^(.+) : warning LNK(\\d+): (.+)$");

			@Override
			public void onStdOutLineWrite(String line) {
				final String exportLibraryPath = libraryPath + ".exp";
				final String importLibraryPath = libraryPath + ".lib";
				final String expectedInfoMessage = String.format("   Creating library %s and object %s", importLibraryPath, exportLibraryPath);
				if (!line.equals(expectedInfoMessage)) {
					Matcher matcher;
					if ((matcher = warningPattern.matcher(line)).find()) {
						final String errorPath = matcher.group(1);
						final int errorNumber = Integer.parseInt(matcher.group(2));
						final String errorMessage = matcher.group(3);
						buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, errorPath, null, errorNumber, errorMessage));
					} else {
						final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
						if (lastBuildMessage == null) {
							System.out.println(line);
						} else {
							lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
						}
					}
				}
			}

			@Override
			public void onStdErrLineWrite(String line) {
				Matcher matcher;
				if ((matcher = fatalErrorPattern.matcher(line)).find()) {
					final int errorNumber = Integer.parseInt(matcher.group(1));
					final String errorMessage = matcher.group(2);
					buildMessages.add(new BuildMessage(BuildMessage.Type.FatalError, null, null, errorNumber, errorMessage));
				} else if ((matcher = warningPattern.matcher(line)).find()) {
					final String errorPath = matcher.group(1);
					final int errorNumber = Integer.parseInt(matcher.group(2));
					final String errorMessage = matcher.group(3);
					buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, errorPath, null, errorNumber, errorMessage));
				} else {
					final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
					if (lastBuildMessage == null) {
						buildMessages.add(new BuildMessage(BuildMessage.Type.Error, null, null, null, line));
					} else {
						lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
					}
				}
			}

			@Override
			public void onError(String description) {
				System.err.println(description);
//					buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, libraryPath, null, null, description));
			}

			@Override
			public void onCompletion(Integer status) {
			}
		};
//			System.out.println(linkerPath + " " + parameters.toString());
		final Process linker = new Process(linkerPath, parameters, environment, eventHandler);
		linker.join();
		return buildMessages;
	}

	private Process.Parameters prepareParameters(String outputPath, Collection<String> objectPaths, boolean isLibrary, String specificationPath) {
		final Process.Parameters parameters = new Process.Parameters();
		parameters.add("/nologo", "/SUBSYSTEM:CONSOLE", "/INCREMENTAL:NO");
		if (isLibrary) {
			parameters.add("/DLL");
			if (specificationPath != null) {
				parameters.add("/DEF:" + specificationPath + ".def");
			}
		}
		parameters.add("/NXCOMPAT");
		if (architecture.equals(Architecture.X86)) {
			parameters.add("/SAFESEH");
		}
		parameters.add("/LARGEADDRESSAWARE");
		if (!isLibrary) {
			parameters.add("/TSAWARE");
		}
		if (!runtimeLibraryUse) {
			parameters.add("/NODEFAULTLIB");
		}
		if (debugEnabled) {
			parameters.add("/DEBUG");
		}
		for (String libraryDirectory : libraryDirectories) {
			parameters.add("/LIBPATH:" + libraryDirectory);
		}
		switch (architecture) {
			case X86:
				parameters.add("/MACHINE:X86"); break;
			case X64:
				parameters.add("/MACHINE:X64"); break;
			case IA64:
				parameters.add("/MACHINE:IA64"); break;
			case ARM:
				parameters.add("/MACHINE:ARM"); break;
		}
		parameters.add("/OUT:" + outputPath);
		for (String objectPath : objectPaths) {
			parameters.add(objectPath);
		}
		for (String library : staticLibraries) {
			parameters.add(library + ".lib");
		}
		for (String library : dynamicLibraries) {
			parameters.add(library + ".lib");
		}
		return parameters;
	}

	private Process.Environment prepareEnvironment() {
		final Process.Environment environment =  new Process.Environment();
		environment.prependPathVariable("PATH", visualStudio.getIdePath());
		environment.prependPathVariable("PATH", visualStudio.getCommonToolsPath());
		environment.prependPathVariable("PATH", visualStudio.getCppToolsPath(architecture));
		for (String globalLibraryDirectory : globalLibraryDirectories) {
			environment.prependPathVariable("LIB", globalLibraryDirectory);
		}
		return environment;
	}

	@Override
	public void addDefaultGlobalLibraryDirectories() {
		this.addGlobalLibraryDirectory(visualStudio.getCppLibraryPath(this.architecture));
		this.addGlobalLibraryDirectory(windowsSdk.getLibraryDirectory(this.architecture));
	}

	@Override
	public Linker copy() {
		return new MicrosoftLinker(this);
	}

	public Architecture getArchitecture() {
		return this.architecture;
	}

	private final Architecture architecture;
	private final VisualStudio visualStudio;
	private final WindowsSdk windowsSdk;
	private final String linkerPath;

}
